# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Literal

import httpx

from ..types import billing_model_create_params, billing_model_update_params
from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from .._utils import (
    maybe_transform,
    async_maybe_transform,
)
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.billing_model import BillingModel
from ..types.billing_model_list_response import BillingModelListResponse

__all__ = ["BillingModelsResource", "AsyncBillingModelsResource"]


class BillingModelsResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> BillingModelsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return BillingModelsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> BillingModelsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return BillingModelsResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        name: str,
        type: Literal["costplus", "subscription", "hybrid"],
        default_price_modifier: Optional[float] | NotGiven = NOT_GIVEN,
        prepaid_amount: Optional[float] | NotGiven = NOT_GIVEN,
        prepaid_max: Optional[float] | NotGiven = NOT_GIVEN,
        threshold: Optional[float] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModel:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/api/v1/billing-model",
            body=maybe_transform(
                {
                    "name": name,
                    "type": type,
                    "default_price_modifier": default_price_modifier,
                    "prepaid_amount": prepaid_amount,
                    "prepaid_max": prepaid_max,
                    "threshold": threshold,
                },
                billing_model_create_params.BillingModelCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModel,
        )

    def retrieve(
        self,
        billing_model_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModel:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not billing_model_id:
            raise ValueError(f"Expected a non-empty value for `billing_model_id` but received {billing_model_id!r}")
        return self._get(
            f"/api/v1/billing-model/{billing_model_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModel,
        )

    def update(
        self,
        billing_model_id: str,
        *,
        type: Literal["costplus", "subscription", "hybrid"],
        default_price_modifier: Optional[float] | NotGiven = NOT_GIVEN,
        name: Optional[str] | NotGiven = NOT_GIVEN,
        prepaid_amount: Optional[float] | NotGiven = NOT_GIVEN,
        prepaid_max: Optional[float] | NotGiven = NOT_GIVEN,
        threshold: Optional[float] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModel:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not billing_model_id:
            raise ValueError(f"Expected a non-empty value for `billing_model_id` but received {billing_model_id!r}")
        return self._put(
            f"/api/v1/billing-model/{billing_model_id}",
            body=maybe_transform(
                {
                    "type": type,
                    "default_price_modifier": default_price_modifier,
                    "name": name,
                    "prepaid_amount": prepaid_amount,
                    "prepaid_max": prepaid_max,
                    "threshold": threshold,
                },
                billing_model_update_params.BillingModelUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModel,
        )

    def list(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModelListResponse:
        return self._get(
            "/api/v1/billing-model",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModelListResponse,
        )

    def set_default(
        self,
        billing_model_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModel:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not billing_model_id:
            raise ValueError(f"Expected a non-empty value for `billing_model_id` but received {billing_model_id!r}")
        return self._put(
            f"/api/v1/billing-model/{billing_model_id}/default",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModel,
        )


class AsyncBillingModelsResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncBillingModelsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return AsyncBillingModelsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncBillingModelsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return AsyncBillingModelsResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        name: str,
        type: Literal["costplus", "subscription", "hybrid"],
        default_price_modifier: Optional[float] | NotGiven = NOT_GIVEN,
        prepaid_amount: Optional[float] | NotGiven = NOT_GIVEN,
        prepaid_max: Optional[float] | NotGiven = NOT_GIVEN,
        threshold: Optional[float] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModel:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/api/v1/billing-model",
            body=await async_maybe_transform(
                {
                    "name": name,
                    "type": type,
                    "default_price_modifier": default_price_modifier,
                    "prepaid_amount": prepaid_amount,
                    "prepaid_max": prepaid_max,
                    "threshold": threshold,
                },
                billing_model_create_params.BillingModelCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModel,
        )

    async def retrieve(
        self,
        billing_model_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModel:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not billing_model_id:
            raise ValueError(f"Expected a non-empty value for `billing_model_id` but received {billing_model_id!r}")
        return await self._get(
            f"/api/v1/billing-model/{billing_model_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModel,
        )

    async def update(
        self,
        billing_model_id: str,
        *,
        type: Literal["costplus", "subscription", "hybrid"],
        default_price_modifier: Optional[float] | NotGiven = NOT_GIVEN,
        name: Optional[str] | NotGiven = NOT_GIVEN,
        prepaid_amount: Optional[float] | NotGiven = NOT_GIVEN,
        prepaid_max: Optional[float] | NotGiven = NOT_GIVEN,
        threshold: Optional[float] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModel:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not billing_model_id:
            raise ValueError(f"Expected a non-empty value for `billing_model_id` but received {billing_model_id!r}")
        return await self._put(
            f"/api/v1/billing-model/{billing_model_id}",
            body=await async_maybe_transform(
                {
                    "type": type,
                    "default_price_modifier": default_price_modifier,
                    "name": name,
                    "prepaid_amount": prepaid_amount,
                    "prepaid_max": prepaid_max,
                    "threshold": threshold,
                },
                billing_model_update_params.BillingModelUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModel,
        )

    async def list(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModelListResponse:
        return await self._get(
            "/api/v1/billing-model",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModelListResponse,
        )

    async def set_default(
        self,
        billing_model_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> BillingModel:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not billing_model_id:
            raise ValueError(f"Expected a non-empty value for `billing_model_id` but received {billing_model_id!r}")
        return await self._put(
            f"/api/v1/billing-model/{billing_model_id}/default",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BillingModel,
        )


class BillingModelsResourceWithRawResponse:
    def __init__(self, billing_models: BillingModelsResource) -> None:
        self._billing_models = billing_models

        self.create = to_raw_response_wrapper(
            billing_models.create,
        )
        self.retrieve = to_raw_response_wrapper(
            billing_models.retrieve,
        )
        self.update = to_raw_response_wrapper(
            billing_models.update,
        )
        self.list = to_raw_response_wrapper(
            billing_models.list,
        )
        self.set_default = to_raw_response_wrapper(
            billing_models.set_default,
        )


class AsyncBillingModelsResourceWithRawResponse:
    def __init__(self, billing_models: AsyncBillingModelsResource) -> None:
        self._billing_models = billing_models

        self.create = async_to_raw_response_wrapper(
            billing_models.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            billing_models.retrieve,
        )
        self.update = async_to_raw_response_wrapper(
            billing_models.update,
        )
        self.list = async_to_raw_response_wrapper(
            billing_models.list,
        )
        self.set_default = async_to_raw_response_wrapper(
            billing_models.set_default,
        )


class BillingModelsResourceWithStreamingResponse:
    def __init__(self, billing_models: BillingModelsResource) -> None:
        self._billing_models = billing_models

        self.create = to_streamed_response_wrapper(
            billing_models.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            billing_models.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            billing_models.update,
        )
        self.list = to_streamed_response_wrapper(
            billing_models.list,
        )
        self.set_default = to_streamed_response_wrapper(
            billing_models.set_default,
        )


class AsyncBillingModelsResourceWithStreamingResponse:
    def __init__(self, billing_models: AsyncBillingModelsResource) -> None:
        self._billing_models = billing_models

        self.create = async_to_streamed_response_wrapper(
            billing_models.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            billing_models.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            billing_models.update,
        )
        self.list = async_to_streamed_response_wrapper(
            billing_models.list,
        )
        self.set_default = async_to_streamed_response_wrapper(
            billing_models.set_default,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Literal, Required, TypedDict

__all__ = ["BillingModelCreateParams"]


class BillingModelCreateParams(TypedDict, total=False):
    name: Required[str]

    type: Required[Literal["costplus", "subscription", "hybrid"]]

    default_price_modifier: Optional[float]

    prepaid_amount: Optional[float]

    prepaid_max: Optional[float]

    threshold: Optional[float]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime
from typing_extensions import Literal

from pydantic import Field as FieldInfo

from .._models import BaseModel
from .total_cost_data import TotalCostData

__all__ = ["PagedLimitList", "Item"]


class Item(BaseModel):
    limit_creation_timestamp: datetime

    limit_id: str

    limit_name: str

    limit_type: Literal["block", "allow"]

    limit_update_timestamp: datetime

    max: float

    totals: TotalCostData

    limit_tags: Optional[List[str]] = None

    threshold: Optional[float] = None


class PagedLimitList(BaseModel):
    current_page: Optional[int] = FieldInfo(alias="currentPage", default=None)

    has_next_page: Optional[bool] = FieldInfo(alias="hasNextPage", default=None)

    has_previous_page: Optional[bool] = FieldInfo(alias="hasPreviousPage", default=None)

    items: Optional[List[Item]] = None

    message: Optional[str] = None

    page_size: Optional[int] = FieldInfo(alias="pageSize", default=None)

    request_id: Optional[str] = None

    total_count: Optional[int] = FieldInfo(alias="totalCount", default=None)

    total_pages: Optional[int] = FieldInfo(alias="totalPages", default=None)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime

from pydantic import Field as FieldInfo

from .._models import BaseModel

__all__ = ["BulkIngestResponse", "Error", "ErrorXproxyResult", "ErrorXproxyResultXproxyError"]


class ErrorXproxyResultXproxyError(BaseModel):
    code: Optional[str] = None

    message: Optional[str] = None


class ErrorXproxyResult(BaseModel):
    message: str

    status_code: int = FieldInfo(alias="statusCode")

    xproxy_error: Optional[ErrorXproxyResultXproxyError] = None


class Error(BaseModel):
    item_index: Optional[int] = None

    xproxy_result: Optional[ErrorXproxyResult] = None
    """
    Represents an generic error that occurred as a result of processing a request.
    APIM returns an (not customizable) error response body of { "statusCode",
    "message" } and this class matches this schema. Derived classes may add
    additional required fields if these classes are specified as produced as a
    return type specific endpoints.
    """


class BulkIngestResponse(BaseModel):
    ingest_count: int

    ingest_timestamp: datetime

    request_id: str

    error_count: Optional[int] = None

    errors: Optional[List[Error]] = None

    total_count: Optional[int] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .shared import PayICommonModelsBudgetManagementCostDetailsBase as PayICommonModelsBudgetManagementCostDetailsBase
from .cost_data import CostData as CostData
from .cost_details import CostDetails as CostDetails
from .billing_model import BillingModel as BillingModel
from .requests_data import RequestsData as RequestsData
from .limit_response import LimitResponse as LimitResponse
from .price_modifier import PriceModifier as PriceModifier
from .ingest_response import IngestResponse as IngestResponse
from .total_cost_data import TotalCostData as TotalCostData
from .default_response import DefaultResponse as DefaultResponse
from .paged_limit_list import PagedLimitList as PagedLimitList
from .category_response import CategoryResponse as CategoryResponse
from .limit_list_params import LimitListParams as LimitListParams
from .ingest_bulk_params import IngestBulkParams as IngestBulkParams
from .ingest_event_param import IngestEventParam as IngestEventParam
from .limit_reset_params import LimitResetParams as LimitResetParams
from .ingest_units_params import IngestUnitsParams as IngestUnitsParams
from .limit_create_params import LimitCreateParams as LimitCreateParams
from .limit_update_params import LimitUpdateParams as LimitUpdateParams
from .bulk_ingest_response import BulkIngestResponse as BulkIngestResponse
from .category_list_response import CategoryListResponse as CategoryListResponse
from .limit_history_response import LimitHistoryResponse as LimitHistoryResponse
from .category_delete_response import CategoryDeleteResponse as CategoryDeleteResponse
from .category_resource_response import CategoryResourceResponse as CategoryResourceResponse
from .billing_model_create_params import BillingModelCreateParams as BillingModelCreateParams
from .billing_model_list_response import BillingModelListResponse as BillingModelListResponse
from .billing_model_update_params import BillingModelUpdateParams as BillingModelUpdateParams
from .experience_instance_response import ExperienceInstanceResponse as ExperienceInstanceResponse
from .price_modifier_create_params import PriceModifierCreateParams as PriceModifierCreateParams
from .price_modifier_update_params import PriceModifierUpdateParams as PriceModifierUpdateParams
from .category_list_resources_response import CategoryListResourcesResponse as CategoryListResourcesResponse
from .price_modifier_retrieve_response import PriceModifierRetrieveResponse as PriceModifierRetrieveResponse
from .category_delete_resource_response import CategoryDeleteResourceResponse as CategoryDeleteResourceResponse
from .pay_i_common_models_api_router_header_info_param import (
    PayICommonModelsAPIRouterHeaderInfoParam as PayICommonModelsAPIRouterHeaderInfoParam,
)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from datetime import datetime

from .._models import BaseModel

__all__ = ["CategoryResponse"]


class CategoryResponse(BaseModel):
    category: str

    resource_count: int

    start_timestamp: datetime



import json
import logging
from typing import Any
from functools import wraps

from wrapt import ObjectProxy, wrap_function_wrapper  # type: ignore

from payi.types.ingest_units_params import Units, IngestUnitsParams
from payi.types.pay_i_common_models_api_router_header_info_param import PayICommonModelsAPIRouterHeaderInfoParam

from .instrument import IsStreaming, PayiInstrumentor


class BedrockInstrumentor:
    @staticmethod
    def instrument(instrumentor: PayiInstrumentor) -> None:
        try:
            import boto3  # type: ignore #  noqa: F401  I001

            wrap_function_wrapper(
                "botocore.client",
                "ClientCreator.create_client",
                create_client_wrapper(instrumentor),
            )

            wrap_function_wrapper(
                "botocore.session",
                "Session.create_client",
                create_client_wrapper(instrumentor),
            )

        except Exception as e:
            logging.debug(f"Error instrumenting bedrock: {e}")
            return

@PayiInstrumentor.payi_wrapper
def create_client_wrapper(instrumentor: PayiInstrumentor, wrapped: Any, instance: Any, *args: Any, **kwargs: Any) -> Any: #  noqa: ARG001
    if kwargs.get("service_name") != "bedrock-runtime":
        return wrapped(*args, **kwargs)

    try:
        client: Any = wrapped(*args, **kwargs)
        client.invoke_model = wrap_invoke(instrumentor, client.invoke_model)
        client.invoke_model_with_response_stream = wrap_invoke_stream(instrumentor, client.invoke_model_with_response_stream)
        client.converse = wrap_converse(instrumentor, client.converse)
        client.converse_stream = wrap_converse_stream(instrumentor, client.converse_stream)

        return client
    except Exception as e:
        logging.debug(f"Error instrumenting bedrock client: {e}")
    
    return wrapped(*args, **kwargs)
    
class InvokeResponseWrapper(ObjectProxy): # type: ignore
    def __init__(
        self,
        response: Any,
        instrumentor: PayiInstrumentor,
        ingest: IngestUnitsParams,
        log_prompt_and_response: bool
        ) -> None:

        super().__init__(response) # type: ignore
        self._response = response
        self._instrumentor = instrumentor
        self._ingest = ingest
        self._log_prompt_and_response = log_prompt_and_response

    def read(self, amt: Any =None): # type: ignore
        # data is array of bytes
        data: Any = self.__wrapped__.read(amt) # type: ignore
        response = json.loads(data)

        resource = self._ingest["resource"]
        if not resource:
            return
        
        input: int = 0
        output: int = 0
        units: dict[str, Units] = self._ingest["units"]

        if resource.startswith("meta.llama3"):
            input = response['prompt_token_count']
            output = response['generation_token_count']
        elif resource.startswith("anthropic."):
            usage = response['usage']
            input = usage['input_tokens']
            output = usage['output_tokens']
        units["text"] = Units(input=input, output=output)

        if self._log_prompt_and_response:
            self._ingest["provider_response_json"] = data.decode('utf-8')
            
        self._instrumentor._ingest_units(self._ingest)

        return data

def wrap_invoke(instrumentor: PayiInstrumentor, wrapped: Any) -> Any:
    @wraps(wrapped)
    def invoke_wrapper(*args: Any, **kwargs: 'dict[str, Any]') -> Any:
        modelId:str = kwargs.get("modelId", "") # type: ignore

        if modelId.startswith("meta.llama3") or modelId.startswith("anthropic."):
            return instrumentor.chat_wrapper(
                "system.aws.bedrock",
                None,
                process_invoke_request,  
                process_synchronous_invoke_response,  
                IsStreaming.false,
                wrapped,
                None,
                args,
                kwargs,
        )
        return wrapped(*args, **kwargs)
    
    return invoke_wrapper

def wrap_invoke_stream(instrumentor: PayiInstrumentor, wrapped: Any) -> Any:
    @wraps(wrapped)
    def invoke_wrapper(*args: Any, **kwargs: Any) -> Any:
        modelId:str = kwargs.get("modelId", "") # type: ignore

        if modelId.startswith("meta.llama3") or modelId.startswith("anthropic."):
            return instrumentor.chat_wrapper(
                "system.aws.bedrock",
                process_invoke_streaming_anthropic_chunk if modelId.startswith("anthropic.") else process_invoke_streaming_llama_chunk,
                process_invoke_request, 
                None,  
                IsStreaming.true,
                wrapped,
                None,
                args,
                kwargs,
            )
        return wrapped(*args, **kwargs)

    return invoke_wrapper

def wrap_converse(instrumentor: PayiInstrumentor, wrapped: Any) -> Any:
    @wraps(wrapped)
    def invoke_wrapper(*args: Any, **kwargs: 'dict[str, Any]') -> Any:
        modelId:str = kwargs.get("modelId", "") # type: ignore

        if modelId.startswith("meta.llama3") or modelId.startswith("anthropic."):
            return instrumentor.chat_wrapper(
                "system.aws.bedrock",
                None,
                process_converse_request,  
                process_synchronous_converse_response,  
                IsStreaming.false,
                wrapped,
                None,
                args,
                kwargs,
        )
        return wrapped(*args, **kwargs)
    
    return invoke_wrapper

def wrap_converse_stream(instrumentor: PayiInstrumentor, wrapped: Any) -> Any:
    @wraps(wrapped)
    def invoke_wrapper(*args: Any, **kwargs: Any) -> Any:
        modelId:str = kwargs.get("modelId", "") # type: ignore

        if modelId.startswith("meta.llama3") or modelId.startswith("anthropic."):
            return instrumentor.chat_wrapper(
                "system.aws.bedrock",
                process_converse_streaming_chunk,
                process_converse_request, 
                None,  
                IsStreaming.true,
                wrapped,
                None,
                args,
                kwargs,
            )
        return wrapped(*args, **kwargs)

    return invoke_wrapper

def process_invoke_streaming_anthropic_chunk(chunk: str, ingest: IngestUnitsParams) -> None:
    chunk_dict =  json.loads(chunk)
    type = chunk_dict.get("type", "")

    if type == "message_start":
        usage = chunk_dict['message']['usage']
        units = ingest["units"]

        input = PayiInstrumentor.update_for_vision(usage['input_tokens'], units)

        units["text"] = Units(input=input, output=0)

        text_cache_write: int = usage.get("cache_creation_input_tokens", 0)
        if text_cache_write > 0:
            units["text_cache_write"] = Units(input=text_cache_write, output=0)

        text_cache_read: int = usage.get("cache_read_input_tokens", 0)
        if text_cache_read > 0:
            units["text_cache_read"] = Units(input=text_cache_read, output=0)

    elif type == "message_delta":
        usage = chunk_dict['usage']
        ingest["units"]["text"]["output"] = usage['output_tokens']

def process_invoke_streaming_llama_chunk(chunk: str, ingest: IngestUnitsParams) -> None:
    chunk_dict =  json.loads(chunk)
    metrics = chunk_dict.get("amazon-bedrock-invocationMetrics", {})
    if metrics:
        input = metrics.get("inputTokenCount", 0)
        output = metrics.get("outputTokenCount", 0)
        ingest["units"]["text"] = Units(input=input, output=output)
    
def process_synchronous_invoke_response(
        response: Any,
        ingest: IngestUnitsParams,
        log_prompt_and_response: bool,
        instrumentor: PayiInstrumentor,
        **kargs: Any) -> Any: #  noqa: ARG001

    metadata = response.get("ResponseMetadata", {})

    request_id = metadata.get("RequestId", "")
    if request_id:
        ingest["provider_response_id"] = request_id

    response_headers = metadata.get("HTTPHeaders", {}).copy()
    if response_headers:
        ingest["provider_response_headers"] = [PayICommonModelsAPIRouterHeaderInfoParam(name=k, value=v) for k, v in response_headers.items()]

    response["body"] = InvokeResponseWrapper(
        response=response["body"],
        instrumentor=instrumentor,
        ingest=ingest,
        log_prompt_and_response=log_prompt_and_response)

    return response

def process_invoke_request(ingest: IngestUnitsParams, *args: Any, **kwargs: Any) -> None: #  noqa: ARG001
    return

def process_converse_streaming_chunk(chunk: 'dict[str, Any]', ingest: IngestUnitsParams) -> None:
    metadata = chunk.get("metadata", {})

    if metadata:
        usage = metadata['usage']
        input = usage["inputTokens"]
        output = usage["outputTokens"]
        ingest["units"]["text"] = Units(input=input, output=output)

def process_synchronous_converse_response(
        response: 'dict[str, Any]',
        ingest: IngestUnitsParams,
        log_prompt_and_response: bool,
        **kargs: Any) -> Any: #  noqa: ARG001

    usage = response["usage"]
    input = usage["inputTokens"]
    output = usage["outputTokens"]
    
    units: dict[str, Units] = ingest["units"]
    units["text"] = Units(input=input, output=output)

    metadata = response.get("ResponseMetadata", {})

    request_id = metadata.get("RequestId", "")
    if request_id:
        ingest["provider_response_id"] = request_id

    response_headers = metadata.get("HTTPHeaders", {})
    if response_headers:
        ingest["provider_response_headers"] = [PayICommonModelsAPIRouterHeaderInfoParam(name=k, value=v) for k, v in response_headers.items()]

    if log_prompt_and_response:
        response_without_metadata = response.copy()
        response_without_metadata.pop("ResponseMetadata", None)
        ingest["provider_response_json"] = json.dumps(response_without_metadata)

    return None    

def process_converse_request(ingest: IngestUnitsParams, *args: Any, **kwargs: Any) -> None: #  noqa: ARG001
    return



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List
from typing_extensions import TypeAlias

from .category_response import CategoryResponse

__all__ = ["CategoryListResponse"]

CategoryListResponse: TypeAlias = List[CategoryResponse]



import json
import logging
from typing import Any, Union
from importlib.metadata import version

import tiktoken  # type: ignore
from wrapt import wrap_function_wrapper  # type: ignore

from payi.types import IngestUnitsParams
from payi.types.ingest_units_params import Units

from .instrument import IsStreaming, PayiInstrumentor


class OpenAiInstrumentor:
    @staticmethod
    def is_azure(instance: Any) -> bool:
        from openai import AzureOpenAI, AsyncAzureOpenAI # type: ignore # noqa: I001

        return isinstance(instance._client, (AsyncAzureOpenAI, AzureOpenAI))

    @staticmethod
    def instrument(instrumentor: PayiInstrumentor) -> None:
        try:
            from openai import OpenAI  # type: ignore #  noqa: F401  I001
            
            wrap_function_wrapper(
                "openai.resources.chat.completions",
                "Completions.create",
                chat_wrapper(instrumentor),
            )

            wrap_function_wrapper(
                "openai.resources.chat.completions",
                "AsyncCompletions.create",
                achat_wrapper(instrumentor),
            )

            wrap_function_wrapper(
                "openai.resources.embeddings",
                "Embeddings.create",
                embeddings_wrapper(instrumentor),
            )

            wrap_function_wrapper(
                "openai.resources.embeddings",
                 "AsyncEmbeddings.create",
                aembeddings_wrapper(instrumentor),
            )

        except Exception as e:
            logging.debug(f"Error instrumenting openai: {e}")
            return


@PayiInstrumentor.payi_wrapper
def embeddings_wrapper(
    instrumentor: PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    return instrumentor.chat_wrapper(
        "system.openai",
        None, # process_chat_chunk,
        None, # process_chat_request,
        process_ebmeddings_synchronous_response,
        IsStreaming.false,
        wrapped,
        instance,
        args,
        kwargs,
    )

@PayiInstrumentor.payi_wrapper
async def aembeddings_wrapper(
    instrumentor: PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    return await instrumentor.achat_wrapper(
        "system.openai",
        None, # process_chat_chunk,
        None, # process_chat_request,
        process_ebmeddings_synchronous_response,
        IsStreaming.false,
        wrapped,
        instance,
        args,
        kwargs,
    )

@PayiInstrumentor.payi_wrapper
def chat_wrapper(
    instrumentor: PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    return instrumentor.chat_wrapper(
        "system.openai",
        process_chat_chunk,
        process_chat_request,
        process_chat_synchronous_response,
        IsStreaming.kwargs,
        wrapped,
        instance,
        args,
        kwargs,
    )

@PayiInstrumentor.payi_awrapper
async def achat_wrapper(
    instrumentor: PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    return await instrumentor.achat_wrapper(
        "system.openai",
        process_chat_chunk,
        process_chat_request,
        process_chat_synchronous_response,
        IsStreaming.kwargs,
        wrapped,
        instance,
        args,
        kwargs,
    )

def process_ebmeddings_synchronous_response(response: str, ingest: IngestUnitsParams, log_prompt_and_response: bool, **kwargs: Any) -> Any: #  noqa: ARG001
    return process_chat_synchronous_response(response, ingest, log_prompt_and_response, **kwargs)

def process_chat_synchronous_response(response: str, ingest: IngestUnitsParams, log_prompt_and_response: bool, **kwargs: Any) -> Any: #  noqa: ARG001
    response_dict = model_to_dict(response)

    add_usage_units(response_dict.get("usage", {}), ingest["units"])

    if log_prompt_and_response:
        ingest["provider_response_json"] = [json.dumps(response_dict)]

    if "id" in response_dict:
        ingest["provider_response_id"] = response_dict["id"]

    return None

def process_chat_chunk(chunk: Any, ingest: IngestUnitsParams) -> None:
    model = model_to_dict(chunk)
    
    if "provider_response_id" not in ingest:
        response_id = model.get("id", None)
        if response_id:
            ingest["provider_response_id"] = response_id

    usage = model.get("usage")
    if usage:
        add_usage_units(usage, ingest["units"])


def model_to_dict(model: Any) -> Any:
    if version("pydantic") < "2.0.0":
        return model.dict()
    if hasattr(model, "model_dump"):
        return model.model_dump()
    elif hasattr(model, "parse"):  # Raw API response
        return model_to_dict(model.parse())
    else:
        return model


def add_usage_units(usage: "dict[str, Any]", units: "dict[str, Units]") -> None:
    input = usage["prompt_tokens"] if "prompt_tokens" in usage else 0
    output = usage["completion_tokens"] if "completion_tokens" in usage else 0
    input_cache = 0

    prompt_tokens_details = usage.get("prompt_tokens_details")
    if prompt_tokens_details:
        input_cache = prompt_tokens_details.get("cached_tokens", 0)
        if input_cache != 0:
            units["text_cache_read"] = Units(input=input_cache, output=0)

    input = PayiInstrumentor.update_for_vision(input - input_cache, units)

    units["text"] = Units(input=input, output=output)

def has_image_and_get_texts(encoding: tiktoken.Encoding, content: Union[str, 'list[Any]']) -> 'tuple[bool, int]':
    if isinstance(content, str):
        return False, 0
    elif isinstance(content, list): # type: ignore
        has_image = any(item.get("type") == "image_url" for item in content)
        if has_image is False:
            return has_image, 0
        
        token_count = sum(len(encoding.encode(item.get("text", ""))) for item in content if item.get("type") == "text")
        return has_image, token_count

def process_chat_request(ingest: IngestUnitsParams, *args: Any, **kwargs: Any) -> None: # noqa: ARG001
    messages = kwargs.get("messages")
    if not messages or len(messages) == 0:
        return
    
    estimated_token_count = 0 
    has_image = False

    try: 
        enc = tiktoken.encoding_for_model(kwargs.get("model")) # type: ignore
    except KeyError:
        enc = tiktoken.get_encoding("o200k_base") # type: ignore
    
    for message in messages:
        msg_has_image, msg_prompt_tokens = has_image_and_get_texts(enc, message.get('content', ''))
        if msg_has_image:
            has_image = True
            estimated_token_count += msg_prompt_tokens
    
    if not has_image or estimated_token_count == 0:
        return

    ingest["units"][PayiInstrumentor.estimated_prompt_tokens] = Units(input=estimated_token_count, output=0)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import httpx

from ..types import price_modifier_create_params, price_modifier_update_params
from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from .._utils import (
    maybe_transform,
    async_maybe_transform,
)
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.price_modifier import PriceModifier
from ..types.price_modifier_retrieve_response import PriceModifierRetrieveResponse

__all__ = ["PriceModifiersResource", "AsyncPriceModifiersResource"]


class PriceModifiersResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> PriceModifiersResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return PriceModifiersResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> PriceModifiersResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return PriceModifiersResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        billing_model_id: str,
        category: str,
        price_modifier: float,
        resource: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PriceModifier:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/api/v1/price-modifier",
            body=maybe_transform(
                {
                    "billing_model_id": billing_model_id,
                    "category": category,
                    "price_modifier": price_modifier,
                    "resource": resource,
                },
                price_modifier_create_params.PriceModifierCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PriceModifier,
        )

    def retrieve(
        self,
        billing_model_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PriceModifierRetrieveResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not billing_model_id:
            raise ValueError(f"Expected a non-empty value for `billing_model_id` but received {billing_model_id!r}")
        return self._get(
            f"/api/v1/price-modifier/{billing_model_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PriceModifierRetrieveResponse,
        )

    def update(
        self,
        *,
        billing_model_id: str,
        category: str,
        price_modifier: float,
        resource: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PriceModifier:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._put(
            "/api/v1/price-modifier",
            body=maybe_transform(
                {
                    "billing_model_id": billing_model_id,
                    "category": category,
                    "price_modifier": price_modifier,
                    "resource": resource,
                },
                price_modifier_update_params.PriceModifierUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PriceModifier,
        )


class AsyncPriceModifiersResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncPriceModifiersResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return AsyncPriceModifiersResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncPriceModifiersResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return AsyncPriceModifiersResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        billing_model_id: str,
        category: str,
        price_modifier: float,
        resource: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PriceModifier:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/api/v1/price-modifier",
            body=await async_maybe_transform(
                {
                    "billing_model_id": billing_model_id,
                    "category": category,
                    "price_modifier": price_modifier,
                    "resource": resource,
                },
                price_modifier_create_params.PriceModifierCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PriceModifier,
        )

    async def retrieve(
        self,
        billing_model_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PriceModifierRetrieveResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not billing_model_id:
            raise ValueError(f"Expected a non-empty value for `billing_model_id` but received {billing_model_id!r}")
        return await self._get(
            f"/api/v1/price-modifier/{billing_model_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PriceModifierRetrieveResponse,
        )

    async def update(
        self,
        *,
        billing_model_id: str,
        category: str,
        price_modifier: float,
        resource: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PriceModifier:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._put(
            "/api/v1/price-modifier",
            body=await async_maybe_transform(
                {
                    "billing_model_id": billing_model_id,
                    "category": category,
                    "price_modifier": price_modifier,
                    "resource": resource,
                },
                price_modifier_update_params.PriceModifierUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PriceModifier,
        )


class PriceModifiersResourceWithRawResponse:
    def __init__(self, price_modifiers: PriceModifiersResource) -> None:
        self._price_modifiers = price_modifiers

        self.create = to_raw_response_wrapper(
            price_modifiers.create,
        )
        self.retrieve = to_raw_response_wrapper(
            price_modifiers.retrieve,
        )
        self.update = to_raw_response_wrapper(
            price_modifiers.update,
        )


class AsyncPriceModifiersResourceWithRawResponse:
    def __init__(self, price_modifiers: AsyncPriceModifiersResource) -> None:
        self._price_modifiers = price_modifiers

        self.create = async_to_raw_response_wrapper(
            price_modifiers.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            price_modifiers.retrieve,
        )
        self.update = async_to_raw_response_wrapper(
            price_modifiers.update,
        )


class PriceModifiersResourceWithStreamingResponse:
    def __init__(self, price_modifiers: PriceModifiersResource) -> None:
        self._price_modifiers = price_modifiers

        self.create = to_streamed_response_wrapper(
            price_modifiers.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            price_modifiers.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            price_modifiers.update,
        )


class AsyncPriceModifiersResourceWithStreamingResponse:
    def __init__(self, price_modifiers: AsyncPriceModifiersResource) -> None:
        self._price_modifiers = price_modifiers

        self.create = async_to_streamed_response_wrapper(
            price_modifiers.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            price_modifiers.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            price_modifiers.update,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime

from .._models import BaseModel

__all__ = ["PriceModifier"]


class PriceModifier(BaseModel):
    billing_model_id: Optional[str] = None

    category: Optional[str] = None

    create_timestamp: datetime

    price_modifier: float

    price_modifier_id: Optional[str] = None

    resource: Optional[str] = None

    resource_id: Optional[str] = None

    update_timestamp: datetime



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["ExperienceType", "LimitConfig"]


class LimitConfig(BaseModel):
    max: float

    limit_tags: Optional[List[str]] = None

    limit_type: Optional[Literal["block", "allow"]] = None

    threshold: Optional[float] = None


class ExperienceType(BaseModel):
    description: str

    name: str

    request_id: str

    limit_config: Optional[LimitConfig] = None

    logging_enabled: Optional[bool] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List, Optional
from typing_extensions import Literal, Required, TypedDict

__all__ = ["TypeCreateParams", "LimitConfig"]


class TypeCreateParams(TypedDict, total=False):
    description: Required[str]

    name: Required[str]

    limit_config: LimitConfig

    logging_enabled: Optional[bool]


class LimitConfig(TypedDict, total=False):
    max: Required[float]

    limit_tags: Optional[List[str]]

    limit_type: Literal["block", "allow"]

    threshold: Optional[float]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from .._models import BaseModel

__all__ = ["ExperienceInstanceResponse"]


class ExperienceInstanceResponse(BaseModel):
    experience_id: str

    limit_id: Optional[str] = None



import json
import uuid
import asyncio
import inspect
import logging
import traceback
from enum import Enum
from typing import Any, Set, Union, Callable, Optional, TypedDict

import nest_asyncio  # type: ignore
from wrapt import ObjectProxy  # type: ignore

from payi import Payi, AsyncPayi
from payi.types import IngestUnitsParams
from payi.lib.helpers import PayiHeaderNames
from payi.types.ingest_response import IngestResponse
from payi.types.ingest_units_params import Units
from payi.types.pay_i_common_models_api_router_header_info_param import PayICommonModelsAPIRouterHeaderInfoParam

from .helpers import PayiCategories
from .Stopwatch import Stopwatch


class Context(TypedDict, total=False):
    proxy: bool
    experience_name: Optional[str]
    experience_id: Optional[str]
    limit_ids: Optional['list[str]']
    request_tags: Optional['list[str]']
    user_id: Optional[str]

class IsStreaming(Enum):
    false = 0
    true = 1 
    kwargs = 2

class PayiInstrumentor:
    estimated_prompt_tokens: str = "estimated_prompt_tokens"

    def __init__(
        self,
        payi: Optional[Payi],
        apayi: Optional[AsyncPayi],
        instruments: Union[Set[str], None] = None,
        log_prompt_and_response: bool = True,
        prompt_and_response_logger: Optional[
            Callable[[str, "dict[str, str]"], None]
        ] = None,  # (request id, dict of data to store) -> None
    ):
        self._payi: Optional[Payi] = payi
        self._apayi: Optional[AsyncPayi] = apayi
        self._context_stack: list[Context] = []  # Stack of context dictionaries
        self._log_prompt_and_response: bool = log_prompt_and_response
        self._prompt_and_response_logger: Optional[Callable[[str, dict[str, str]], None]] = prompt_and_response_logger

        self._blocked_limits: set[str] = set()
        self._exceeded_limits: set[str] = set()

        if instruments is None or "*" in instruments:
            self._instrument_all()
        else:
            self._instrument_specific(instruments)

    def _instrument_all(self) -> None:
        self._instrument_openai()
        self._instrument_anthropic()
        self._instrument_aws_bedrock()

    def _instrument_specific(self, instruments: Set[str]) -> None:
        if PayiCategories.openai in instruments or PayiCategories.azure_openai in instruments:
            self._instrument_openai()
        if PayiCategories.anthropic in instruments:
            self._instrument_anthropic()
        if PayiCategories.aws_bedrock in instruments:
            self._instrument_aws_bedrock()

    def _instrument_openai(self) -> None:
        from .OpenAIInstrumentor import OpenAiInstrumentor

        try:
            OpenAiInstrumentor.instrument(self)

        except Exception as e:
            logging.error(f"Error instrumenting OpenAI: {e}")

    def _instrument_anthropic(self) -> None:
        from .AnthropicInstrumentor import AnthropicIntrumentor

        try:
            AnthropicIntrumentor.instrument(self)

        except Exception as e:
            logging.error(f"Error instrumenting Anthropic: {e}")

    def _instrument_aws_bedrock(self) -> None:
        from .BedrockInstrumentor import BedrockInstrumentor

        try:
            BedrockInstrumentor.instrument(self)

        except Exception as e:
            logging.error(f"Error instrumenting AWS bedrock: {e}")

    def _process_ingest_units(self, ingest_units: IngestUnitsParams, log_data: 'dict[str, str]') -> bool:
        if int(ingest_units.get("http_status_code") or 0) < 400:
            units = ingest_units.get("units", {})
            if not units or all(unit.get("input", 0) == 0 and unit.get("output", 0) == 0 for unit in units.values()):
                logging.error(
                    'No units to ingest.  For OpenAI streaming calls, make sure you pass stream_options={"include_usage": True}'
                )
                return False

        if self._log_prompt_and_response and self._prompt_and_response_logger:
            response_json = ingest_units.pop("provider_response_json", None)
            request_json = ingest_units.pop("provider_request_json", None)
            stack_trace = ingest_units.get("properties", {}).pop("system.stack_trace", None)  # type: ignore

            if response_json is not None:
                # response_json is a list of strings, convert a single json string
                log_data["provider_response_json"] = json.dumps(response_json)
            if request_json is not None:
                log_data["provider_request_json"] = request_json
            if stack_trace is not None:
                log_data["stack_trace"] = stack_trace

        return True

    def _process_ingest_units_response(self, ingest_response: IngestResponse) -> None:
        if ingest_response.xproxy_result.limits:
            for limit_id, state in ingest_response.xproxy_result.limits.items():
                removeBlockedId: bool = False

                if state.state == "blocked":
                    self._blocked_limits.add(limit_id)
                elif state.state == "exceeded":
                    self._exceeded_limits.add(limit_id)
                    removeBlockedId = True
                elif state.state == "ok":
                    removeBlockedId = True

                # opportunistically remove blocked limits
                if removeBlockedId:
                    self._blocked_limits.discard(limit_id)

    async def _aingest_units(self, ingest_units: IngestUnitsParams) -> Optional[IngestResponse]:
        ingest_response: Optional[IngestResponse] = None
    
        # return early if there are no units to ingest and on a successul ingest request
        log_data: 'dict[str,str]' = {}
        if not self._process_ingest_units(ingest_units, log_data):
            return None

        try:
            if self._apayi:    
                ingest_response= await self._apayi.ingest.units(**ingest_units)
            elif self._payi:
                ingest_response = self._payi.ingest.units(**ingest_units)
            else:
                logging.error("No payi instance to ingest units")
                return None         

            if ingest_response:
                self._process_ingest_units_response(ingest_response)

                if ingest_response and self._log_prompt_and_response and self._prompt_and_response_logger:
                    request_id = ingest_response.xproxy_result.request_id
                    self._prompt_and_response_logger(request_id, log_data)  # type: ignore

            return ingest_response
        except Exception as e:
            logging.error(f"Error Pay-i ingesting result: {e}")
    
        return None         
    
    def _call_aingest_sync(self, ingest_units: IngestUnitsParams) -> Optional[IngestResponse]:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None
        
        try:
            if loop and loop.is_running():
                nest_asyncio.apply(loop) # type: ignore
                return asyncio.run(self._aingest_units(ingest_units))                
            else:
                # When there's no running loop, create a new one
                return asyncio.run(self._aingest_units(ingest_units))
        except Exception as e:
            logging.error(f"Error calling aingest_units synchronously: {e}")
        return None
        
    def _ingest_units(self, ingest_units: IngestUnitsParams) -> Optional[IngestResponse]:
        ingest_response: Optional[IngestResponse] = None
        
        # return early if there are no units to ingest and on a successul ingest request
        log_data: 'dict[str,str]' = {}
        if not self._process_ingest_units(ingest_units, log_data):
            return None

        try:
            if self._payi:
                ingest_response = self._payi.ingest.units(**ingest_units)
                self._process_ingest_units_response(ingest_response)

                if self._log_prompt_and_response and self._prompt_and_response_logger:
                    request_id = ingest_response.xproxy_result.request_id
                    self._prompt_and_response_logger(request_id, log_data)  # type: ignore

                return ingest_response
            elif self._apayi:
                # task runs async. aingest_units will invoke the callback and post process
                ingest_response = self._call_aingest_sync(ingest_units)
                return ingest_response
            else:
                logging.error("No payi instance to ingest units")

        except Exception as e:
            logging.error(f"Error Pay-i ingesting result: {e}")
        
        return None

    def _setup_call_func(
        self
        ) -> 'tuple[Context, Optional[str], Optional[str]]':
        context: Context = {}

        if len(self._context_stack) > 0:
            # copy current context into the upcoming context
            context = self._context_stack[-1].copy()
            context.pop("proxy")
            previous_experience_name = context.get("experience_name", None)
            previous_experience_id = context.get("experience_id", None)
        else:
            previous_experience_name = None
            previous_experience_id = None
        return (context, previous_experience_name, previous_experience_id)

    def _init_context(
        self,
        context: Context,
        previous_experience_name: Optional[str],
        previous_experience_id: Optional[str],
        proxy: bool,
        limit_ids: Optional["list[str]"],
        request_tags: Optional["list[str]"],
        experience_name: Optional[str],
        experience_id: Optional[str],
        user_id: Optional[str],
        ) -> None:
        context["proxy"] = proxy

        # Handle experience name and ID logic
        if not experience_name:
            # If no experience_name specified, use previous values
            context["experience_name"] = previous_experience_name
            context["experience_id"] = previous_experience_id
        else:
            # If experience_name is specified
            if experience_name == previous_experience_name:
                # Same experience name, use previous ID unless new one specified
                context["experience_name"] = experience_name
                context["experience_id"] = experience_id if experience_id else previous_experience_id
            else:
                # Different experience name, use specified ID or generate one
                context["experience_name"] = experience_name
                context["experience_id"] = experience_id if experience_id else str(uuid.uuid4())

        # set any values explicitly passed by the caller, otherwise use what is already in the context
        if limit_ids:
            context["limit_ids"] = limit_ids
        if request_tags:
            context["request_tags"] = request_tags
        if user_id:
            context["user_id"] = user_id

        self.set_context(context)
        
    async def _acall_func(
        self,
        func: Any,
        proxy: bool,
        limit_ids: Optional["list[str]"],
        request_tags: Optional["list[str]"],
        experience_name: Optional[str],
        experience_id: Optional[str],
        user_id: Optional[str],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        context, previous_experience_name, previous_experience_id = self._setup_call_func()

        with self:
            self._init_context(
                context,
                previous_experience_name,
                previous_experience_id,
                proxy, 
                limit_ids, 
                request_tags,
                experience_name,
                experience_id,
                user_id)
            return await func(*args, **kwargs)

    def _call_func(
        self,
        func: Any,
        proxy: bool,
        limit_ids: Optional["list[str]"],
        request_tags: Optional["list[str]"],
        experience_name: Optional[str],
        experience_id: Optional[str],
        user_id: Optional[str],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        context, previous_experience_name, previous_experience_id = self._setup_call_func()

        with self:
            self._init_context(
                context,
                previous_experience_name,
                previous_experience_id,
                proxy, 
                limit_ids, 
                request_tags,
                experience_name,
                experience_id,
                user_id)
            return func(*args, **kwargs)

    def __enter__(self) -> Any:
        # Push a new context dictionary onto the stack
        self._context_stack.append({})
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        # Pop the current context off the stack
        if self._context_stack:
            self._context_stack.pop()

    def set_context(self, context: Context) -> None:
        # Update the current top of the stack with the provided context
        if self._context_stack:
            self._context_stack[-1].update(context)

    def get_context(self) -> Optional[Context]:
        # Return the current top of the stack
        return self._context_stack[-1] if self._context_stack else None


    def _prepare_ingest(
        self,
        ingest: IngestUnitsParams,
        ingest_extra_headers: "dict[str, str]", # do not coflict potential kwargs["extra_headers"]
        **kwargs: Any,
    ) -> None:
        limit_ids = ingest_extra_headers.pop(PayiHeaderNames.limit_ids, None)
        request_tags = ingest_extra_headers.pop(PayiHeaderNames.request_tags, None)
        experience_name = ingest_extra_headers.pop(PayiHeaderNames.experience_name, None)
        experience_id = ingest_extra_headers.pop(PayiHeaderNames.experience_id, None)
        user_id = ingest_extra_headers.pop(PayiHeaderNames.user_id, None)

        if limit_ids:
            ingest["limit_ids"] = limit_ids.split(",")
        if request_tags:
            ingest["request_tags"] = request_tags.split(",")
        if experience_name:
            ingest["experience_name"] = experience_name
        if experience_id:
            ingest["experience_id"] = experience_id
        if user_id:
            ingest["user_id"] = user_id

        if len(ingest_extra_headers) > 0:
            ingest["provider_request_headers"] = [PayICommonModelsAPIRouterHeaderInfoParam(name=k, value=v) for k, v in ingest_extra_headers.items()]

        provider_prompt: "dict[str, Any]" = {}
        for k, v in kwargs.items():
            if k == "messages":
                provider_prompt[k] = [m.model_dump() if hasattr(m, "model_dump") else m for m in v]
            elif k in ["extra_headers", "extra_query"]:
                pass
            else:
                try:
                    json.dumps(v)
                    provider_prompt[k] = v
                except (TypeError, ValueError):
                    pass

        if self._log_prompt_and_response:
            ingest["provider_request_json"] = json.dumps(provider_prompt)

    async def achat_wrapper(
        self,
        category: str,
        process_chunk: Optional[Callable[[Any, IngestUnitsParams], None]],
        process_request: Optional[Callable[[IngestUnitsParams, Any, Any], None]],
        process_synchronous_response: Any,
        is_streaming: IsStreaming,
        wrapped: Any,
        instance: Any,
        args: Any,
        kwargs: Any,
    ) -> Any:
        context = self.get_context()

        # Bedrock client does not have an async method

        if not context:   
            # wrapped function invoked outside of decorator scope
            return await wrapped(*args, **kwargs)

        # after _udpate_headers, all metadata to add to ingest is in extra_headers, keyed by the xproxy-xxx header name
        extra_headers = kwargs.get("extra_headers", {})
        self._update_headers(context, extra_headers)

        if context.get("proxy", True):
            if "extra_headers" not in kwargs:
                kwargs["extra_headers"] = extra_headers

            return await wrapped(*args, **kwargs)

        ingest: IngestUnitsParams = {"category": category, "units": {}} # type: ignore
        ingest["resource"] = kwargs.get("model", "")

        if category == PayiCategories.openai and instance and hasattr(instance, "_client"):
            from .OpenAIInstrumentor import OpenAiInstrumentor # noqa: I001

            if OpenAiInstrumentor.is_azure(instance):
                route_as_resource = extra_headers.pop(PayiHeaderNames.route_as_resource, None)
                resource_scope = extra_headers.pop(PayiHeaderNames.resource_scope, None)

                if not route_as_resource:
                    logging.error("Azure OpenAI route as resource not found, not ingesting")
                    return await wrapped(*args, **kwargs)

                if resource_scope:
                    if not(resource_scope in ["global", "datazone"] or resource_scope.startswith("region")):
                        logging.error("Azure OpenAI invalid resource scope, not ingesting")
                        return wrapped(*args, **kwargs)

                    ingest["resource_scope"] = resource_scope

                category = PayiCategories.azure_openai

                ingest["category"] = category
                ingest["resource"] = route_as_resource

        current_frame = inspect.currentframe()
        # f_back excludes the current frame, strip() cleans up whitespace and newlines
        stack = [frame.strip() for frame in traceback.format_stack(current_frame.f_back)]  # type: ignore

        ingest['properties'] = { 'system.stack_trace': json.dumps(stack) }

        if process_request:
            process_request(ingest, (), instance)

        sw = Stopwatch()
        stream: bool = False
        
        if is_streaming == IsStreaming.kwargs:
            stream = kwargs.get("stream", False)
        elif is_streaming == IsStreaming.true:
            stream = True
        else:
            stream = False

        try:
            self._prepare_ingest(ingest, extra_headers, **kwargs)
            sw.start()
            response = await wrapped(*args, **kwargs)

        except Exception as e:  # pylint: disable=broad-except
            sw.stop()
            duration = sw.elapsed_ms_int()

            # TODO ingest error

            raise e

        if stream:
            stream_result = ChatStreamWrapper(
                response=response,
                instance=instance,
                instrumentor=self,
                log_prompt_and_response=self._log_prompt_and_response,
                ingest=ingest,
                stopwatch=sw,
                process_chunk=process_chunk,
                is_bedrock=False,
            )

            return stream_result

        sw.stop()
        duration = sw.elapsed_ms_int()
        ingest["end_to_end_latency_ms"] = duration
        ingest["http_status_code"] = 200

        if process_synchronous_response:
            return_result: Any = process_synchronous_response(
                response=response,
                ingest=ingest,
                log_prompt_and_response=self._log_prompt_and_response,
                instrumentor=self)
            if return_result:
                return return_result

        await self._aingest_units(ingest)

        return response

    def chat_wrapper(
        self,
        category: str,
        process_chunk: Optional[Callable[[Any, IngestUnitsParams], None]],
        process_request: Optional[Callable[[IngestUnitsParams, Any, Any], None]],
        process_synchronous_response: Any,
        is_streaming: IsStreaming,
        wrapped: Any,
        instance: Any,
        args: Any,
        kwargs: Any,
    ) -> Any:
        context = self.get_context()

        is_bedrock:bool = category == PayiCategories.aws_bedrock

        if not context:
            if is_bedrock:
                # boto3 doesn't allow extra_headers
                kwargs.pop("extra_headers", None)
    
            # wrapped function invoked outside of decorator scope
            return wrapped(*args, **kwargs)

        # after _udpate_headers, all metadata to add to ingest is in extra_headers, keyed by the xproxy-xxx header name
        extra_headers = kwargs.get("extra_headers", {})
        self._update_headers(context, extra_headers)

        if context.get("proxy", True):
            if "extra_headers" not in kwargs:
                kwargs["extra_headers"] = extra_headers

            return wrapped(*args, **kwargs)
        
        ingest: IngestUnitsParams = {"category": category, "units": {}} # type: ignore
        if is_bedrock:
            # boto3 doesn't allow extra_headers
            kwargs.pop("extra_headers", None)
            ingest["resource"] = kwargs.get("modelId", "")
        else:
            ingest["resource"] = kwargs.get("model", "")

        if category == PayiCategories.openai and instance and hasattr(instance, "_client"):
            from .OpenAIInstrumentor import OpenAiInstrumentor # noqa: I001

            if OpenAiInstrumentor.is_azure(instance):
                route_as_resource:str = extra_headers.pop(PayiHeaderNames.route_as_resource, None)
                resource_scope:str = extra_headers.pop(PayiHeaderNames.resource_scope, None)

                if not route_as_resource:
                    logging.error("Azure OpenAI route as resource not found, not ingesting")
                    return wrapped(*args, **kwargs)

                if resource_scope:
                    if not(resource_scope in ["global", "datazone"] or resource_scope.startswith("region")):
                        logging.error("Azure OpenAI invalid resource scope, not ingesting")
                        return wrapped(*args, **kwargs)

                    ingest["resource_scope"] = resource_scope

                category = PayiCategories.azure_openai

                ingest["category"] = category
                ingest["resource"] = route_as_resource

        current_frame = inspect.currentframe()
        # f_back excludes the current frame, strip() cleans up whitespace and newlines
        stack = [frame.strip() for frame in traceback.format_stack(current_frame.f_back)]  # type: ignore

        ingest['properties'] = { 'system.stack_trace': json.dumps(stack) }

        if process_request:
            process_request(ingest, (), kwargs)

        sw = Stopwatch()
        stream: bool = False
        
        if is_streaming == IsStreaming.kwargs:
            stream = kwargs.get("stream", False)
        elif is_streaming == IsStreaming.true:
            stream = True
        else:
            stream = False

        try:
            self._prepare_ingest(ingest, extra_headers, **kwargs)
            sw.start()
            response = wrapped(*args, **kwargs)

        except Exception as e:  # pylint: disable=broad-except
            sw.stop()
            duration = sw.elapsed_ms_int()

            # TODO ingest error

            raise e

        if stream:
            stream_result = ChatStreamWrapper(
                response=response,
                instance=instance,
                instrumentor=self,
                log_prompt_and_response=self._log_prompt_and_response,
                ingest=ingest,
                stopwatch=sw,
                process_chunk=process_chunk,
                is_bedrock=is_bedrock,
            )

            if is_bedrock:
                if "body" in response:
                    response["body"] = stream_result
                else:
                    response["stream"] = stream_result
                return response
            
            return stream_result

        sw.stop()
        duration = sw.elapsed_ms_int()
        ingest["end_to_end_latency_ms"] = duration
        ingest["http_status_code"] = 200

        if process_synchronous_response:
            return_result: Any = process_synchronous_response(
                response=response,
                ingest=ingest,
                log_prompt_and_response=self._log_prompt_and_response,
                instrumentor=self)
            if return_result:
                return return_result

        self._ingest_units(ingest)

        return response

    @staticmethod
    def _update_headers(
        context: Context,
        extra_headers: "dict[str, str]",
    ) -> None:
        limit_ids: Optional[list[str]] = context.get("limit_ids")
        request_tags: Optional[list[str]] = context.get("request_tags")
        experience_name: Optional[str] = context.get("experience_name")
        experience_id: Optional[str] = context.get("experience_id")
        user_id: Optional[str] = context.get("user_id")

        # Merge limits from the decorator and extra headers
        if limit_ids is not None:
            existing_limit_ids = extra_headers.get(PayiHeaderNames.limit_ids, None)
            
            if not existing_limit_ids:
                extra_headers[PayiHeaderNames.limit_ids] = ",".join(limit_ids)
            else:
                existing_ids = existing_limit_ids.split(',')
                combined_ids = list(set(existing_ids + limit_ids))
                extra_headers[PayiHeaderNames.limit_ids] = ",".join(combined_ids)

        # Merge request from the decorator and extra headers
        if request_tags is not None:
            existing_request_tags = extra_headers.get(PayiHeaderNames.request_tags, None)

            if not existing_request_tags:
                extra_headers[PayiHeaderNames.request_tags] = ",".join(request_tags)
            else:
                existing_tags = existing_request_tags.split(',')
                combined_tags = list(set(existing_tags + request_tags))
                extra_headers[PayiHeaderNames.request_tags] = ",".join(combined_tags)

        # inner extra_headers user_id takes precedence over outer decorator user_id
        if user_id is not None and extra_headers.get(PayiHeaderNames.user_id, None) is None:
            extra_headers[PayiHeaderNames.user_id] = user_id   

        # inner extra_headers experience_name and experience_id take precedence over outer decorator experience_name and experience_id
        # if either inner value is specified, ignore outer decorator values
        if PayiHeaderNames.experience_name not in extra_headers and PayiHeaderNames.experience_id not in extra_headers:

            # use both decorator values
            if experience_name is not None:
                extra_headers[PayiHeaderNames.experience_name] = experience_name
            if experience_id is not None:
                extra_headers[PayiHeaderNames.experience_id] = experience_id
        
        elif PayiHeaderNames.experience_id in extra_headers and PayiHeaderNames.experience_name not in extra_headers:
            # use the decorator experience name and the inner experience id
            if experience_name is not None:
                extra_headers[PayiHeaderNames.experience_name] = experience_name
        
        else:
            # use the inner experience name and id as-is
            ...

    @staticmethod
    def update_for_vision(input: int, units: 'dict[str, Units]') -> int:
        if PayiInstrumentor.estimated_prompt_tokens in units:
            prompt_token_estimate: int = units.pop(PayiInstrumentor.estimated_prompt_tokens)["input"] # type: ignore
            vision = input - prompt_token_estimate
            if (vision > 0):
                units["vision"] = Units(input=vision, output=0)
                input = prompt_token_estimate
        
        return input

    @staticmethod
    def payi_wrapper(func: Any) -> Any:
        def _payi_wrapper(o: Any) -> Any:
            def wrapper(wrapped: Any, instance: Any, args: Any, kwargs: Any) -> Any:
                return func(
                    o,
                    wrapped,
                    instance,
                    *args,
                    **kwargs,
                )

            return wrapper

        return _payi_wrapper

    @staticmethod
    def payi_awrapper(func: Any) -> Any:
        def _payi_awrapper(o: Any) -> Any:
            async def wrapper(wrapped: Any, instance: Any, args: Any, kwargs: Any) -> Any:
                return await func(
                    o,
                    wrapped,
                    instance,
                    *args,
                    **kwargs,
                )

            return wrapper

        return _payi_awrapper

class ChatStreamWrapper(ObjectProxy):  # type: ignore
    def __init__(
        self,
        response: Any,
        instance: Any,
        instrumentor: PayiInstrumentor,
        ingest: IngestUnitsParams,
        stopwatch: Stopwatch,
        process_chunk: Optional[Callable[[Any, IngestUnitsParams], None]] = None,
        log_prompt_and_response: bool = True,
        is_bedrock: bool = False,
    ) -> None:

        bedrock_from_stream: bool = False
        if is_bedrock:
            ingest["provider_response_id"] = response["ResponseMetadata"]["RequestId"]
            stream = response.get("stream", None)

            if stream:
                response = stream
                bedrock_from_stream = True
            else:
                response = response.get("body")
                bedrock_from_stream = False

        super().__init__(response)  # type: ignore

        self._response = response
        self._instance = instance

        self._instrumentor = instrumentor
        self._stopwatch: Stopwatch = stopwatch
        self._ingest: IngestUnitsParams = ingest
        self._log_prompt_and_response: bool = log_prompt_and_response
        self._responses: list[str] = []

        self._process_chunk: Optional[Callable[[Any, IngestUnitsParams], None]] = process_chunk

        self._first_token: bool = True
        self._is_bedrock: bool = is_bedrock
        self._bedrock_from_stream: bool = bedrock_from_stream

    def __enter__(self) -> Any:
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None: 
        self.__wrapped__.__exit__(exc_type, exc_val, exc_tb)  # type: ignore

    async def __aenter__(self) -> Any:
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.__wrapped__.__aexit__(exc_type, exc_val, exc_tb)  # type: ignore

    def __iter__(self) -> Any:  
        if self._is_bedrock:
            # MUST be reside in a separate function so that the yield statement doesn't implicitly return its own iterator and overriding self
            return self._iter_bedrock()
        return self

    def _iter_bedrock(self) -> Any:
        # botocore EventStream doesn't have a __next__ method so iterate over the wrapped object in place
        for event in self.__wrapped__: # type: ignore
            if (self._bedrock_from_stream):
                self._evaluate_chunk(event)
            else:
                chunk = event.get('chunk') # type: ignore
                if chunk:
                    decode = chunk.get('bytes').decode() # type: ignore
                    self._evaluate_chunk(decode)
            yield event

        self._stop_iteration()

    def __aiter__(self) -> Any:
        return self

    def __next__(self) -> Any:
        try:
            chunk: Any = self.__wrapped__.__next__()  # type: ignore
        except Exception as e:
            if isinstance(e, StopIteration):
                self._stop_iteration()
            raise e
        else:
            self._evaluate_chunk(chunk)
            return chunk

    async def __anext__(self) -> Any:
        try:
            chunk: Any = await self.__wrapped__.__anext__()  # type: ignore
        except Exception as e:
            if isinstance(e, StopAsyncIteration):
                await self._astop_iteration()
            raise e
        else:
            self._evaluate_chunk(chunk)
            return chunk

    def _evaluate_chunk(self, chunk: Any) -> None:
        if self._first_token:
            self._ingest["time_to_first_token_ms"] = self._stopwatch.elapsed_ms_int()
            self._first_token = False

        if self._log_prompt_and_response:
            self._responses.append(self.chunk_to_json(chunk))

        if self._process_chunk:
            self._process_chunk(chunk, self._ingest)

    def _process_stop_iteration(self) -> None:
        self._stopwatch.stop()
        self._ingest["end_to_end_latency_ms"] = self._stopwatch.elapsed_ms_int()
        self._ingest["http_status_code"] = 200

        if self._log_prompt_and_response:
            self._ingest["provider_response_json"] = self._responses

    async def _astop_iteration(self) -> None:
        self._process_stop_iteration()
        await self._instrumentor._aingest_units(self._ingest)

    def _stop_iteration(self) -> None:
        self._process_stop_iteration()
        self._instrumentor._ingest_units(self._ingest)

    @staticmethod
    def chunk_to_json(chunk: Any) -> str:
        if hasattr(chunk, "to_json"):
            return str(chunk.to_json())
        elif isinstance(chunk, bytes):
            return chunk.decode()
        elif isinstance(chunk, str):
            return chunk
        else:
            # assume dict
            return json.dumps(chunk)

global _instrumentor
_instrumentor: Optional[PayiInstrumentor] = None

def payi_instrument(
    payi: Optional[Union[Payi, AsyncPayi, 'list[Union[Payi, AsyncPayi]]']] = None,
    instruments: Optional[Set[str]] = None,
    log_prompt_and_response: bool = True,
    prompt_and_response_logger: Optional[Callable[[str, "dict[str, str]"], None]] = None,
) -> None:
    global _instrumentor
    if _instrumentor:
        return
    
    payi_param: Optional[Payi] = None
    apayi_param: Optional[AsyncPayi] = None

    if isinstance(payi, Payi):
        payi_param = payi
    elif isinstance(payi, AsyncPayi):
        apayi_param = payi
    elif isinstance(payi, list):
        for p in payi:
            if isinstance(p, Payi):
                payi_param = p
            elif isinstance(p, AsyncPayi): # type: ignore
                apayi_param = p

    # allow for both payi and apayi to be None for the @proxy case
    
    _instrumentor = PayiInstrumentor(
        payi=payi_param,
        apayi=apayi_param,
        instruments=instruments,
        log_prompt_and_response=log_prompt_and_response,
        prompt_and_response_logger=prompt_and_response_logger,
    )

def ingest(
    limit_ids: Optional["list[str]"] = None,
    request_tags: Optional["list[str]"] = None,
    experience_name: Optional[str] = None,
    experience_id: Optional[str] = None,
    user_id: Optional[str] = None,
) -> Any:
    def _ingest(func: Any) -> Any:
        import asyncio
        if asyncio.iscoroutinefunction(func):
            async def awrapper(*args: Any, **kwargs: Any) -> Any:
                if not _instrumentor:
                    return await func(*args, **kwargs)
                # Call the instrumentor's _call_func for async functions
                return await _instrumentor._acall_func(
                    func,
                    False,
                    limit_ids,
                    request_tags,
                    experience_name,
                    experience_id,
                    user_id,
                    *args,
                    **kwargs,
                )
            return awrapper
        else:
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                if not _instrumentor:
                    return func(*args, **kwargs)
                return _instrumentor._call_func(
                    func,
                    False,
                    limit_ids,
                    request_tags,
                    experience_name,
                    experience_id,
                    user_id,
                    *args,
                    **kwargs,
                )
            return wrapper
    return _ingest

def proxy(
    limit_ids: Optional["list[str]"] = None,
    request_tags: Optional["list[str]"] = None,
    experience_name: Optional[str] = None,
    experience_id: Optional[str] = None,
    user_id: Optional[str] = None,
) -> Any:
    def _proxy(func: Any) -> Any:
        import asyncio
        if asyncio.iscoroutinefunction(func):
            async def _proxy_awrapper(*args: Any, **kwargs: Any) -> Any:
                if not _instrumentor:
                    return await func(*args, **kwargs)
                return await _instrumentor._call_func(
                    func,
                    True,
                    limit_ids,
                    request_tags,
                    experience_name,
                    experience_id,
                    user_id,
                    *args,
                    **kwargs
                )

            return _proxy_awrapper
        else:
            def _proxy_wrapper(*args: Any, **kwargs: Any) -> Any:
                if not _instrumentor:
                    return func(*args, **kwargs)
                return _instrumentor._call_func(
                    func,
                    True,
                    limit_ids,
                    request_tags,
                    experience_name,
                    experience_id,
                    user_id,
                    *args,
                    **kwargs
                )

            return _proxy_wrapper

    return _proxy



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.


from .._models import BaseModel
from .cost_details import CostDetails
from .shared.pay_i_common_models_budget_management_cost_details_base import (
    PayICommonModelsBudgetManagementCostDetailsBase,
)

__all__ = ["CostData"]


class CostData(BaseModel):
    input: PayICommonModelsBudgetManagementCostDetailsBase

    output: PayICommonModelsBudgetManagementCostDetailsBase

    total: CostDetails
